import socket as socket_module
from collections.abc import Iterator
from pathlib import Path

from doip_sdk.constant import ResponseStatus
from doip_sdk.model.response import DOIPResponse


def write_string_segment(socket: socket_module.socket, message: str, encoding: str = 'utf-8'):
    message_byte = bytes(message, encoding)
    socket.sendall(message_byte + b'\n#\n')


def write_file_segment(socket: socket_module.socket, file: Path, chunk_size: int = 1024):
    socket.sendall(b'@\n')
    with file.open('rb') as f:
        while bytes_read := f.read(chunk_size):
            chunk_size_byte_string = bytes(f'{len(bytes_read)}\n', encoding='utf-8')
            socket.sendall(chunk_size_byte_string)
            socket.sendall(bytes_read + b'\n')
    socket.sendall(b'#\n')


def write_empty_segment(socket: socket_module.socket):
    socket.sendall(b'#\n')


def send_unknown_operation_response(socket: socket_module.socket, request_id: str | None = None):
    output = {
        'reason': 'This operation is not supported.'
    }
    response = DOIPResponse(requestId=request_id, status=ResponseStatus.UNKNOWN_OPERATION, output=output)
    write_string_segment(socket=socket, message=response.model_dump_json(exclude_none=True))
    write_empty_segment(socket=socket)


def send_invalid_request_response(socket: socket_module.socket, reason: str, request_id: str | None = None):
    output = {
        'reason': reason
    }
    response = DOIPResponse(requestId=request_id, status=ResponseStatus.INVALID, output=output)
    write_string_segment(socket=socket, message=response.model_dump_json(exclude_none=True))
    write_empty_segment(socket=socket)


class SocketReader:

    def __init__(self, socket: socket_module.socket):
        self.socket = socket
        self.rfile = None

    def get_chunks(self) -> Iterator[bytearray]:
        """
        This method parses the socket data stream, which follows DOIP Specification, and returns it chunk by chunk.
        A chunk can either be:

        1. a JSON segment, or
        2. an ``@`` sign, which indicates next chunks are chunks in a byte segment, or
        3. an ``#`` sign, which indicates the end of a byte segment.

        Returns
        -------
        A `generator <https://wiki.python.org/moin/Generators>`_, where each element is a chunk described above.
        """
        try:
            self.rfile = self.socket.makefile('rb', -1)
            buffer = bytearray()
            while True:
                received = self.rfile.readline().strip()
                match received:
                    case b'#':
                        # Reading a # when the buffer is empty means that it is an empty segment, which indicates the
                        # end of the data stream -> finish reading
                        if not buffer:
                            break

                        # Otherwise, yield the JSON segment and reset the buffer for the next read.
                        yield buffer
                        buffer = bytearray()
                    case b'@':
                        # Yield the @ sign to indicate the start of a byte segment
                        yield received

                        # Yield chunks in the byte segment
                        while file_chunk := self._read_chunk_in_file_segment():
                            yield file_chunk

                        # Yield the # sign to indicate the end of a byte segment
                        yield b'#'
                    case _:
                        buffer.extend(received)
        finally:
            if self.rfile:
                self.rfile.close()

    def _read_chunk_in_file_segment(self):
        received = self.rfile.readline().strip()

        # `received` can be empty because of the newline character at the end of a byte chunk. This character is not
        # counted toward the length of the byte chunk. Therefore, just ignore it and read further
        if not received:
            received = self.rfile.readline().strip()

        # End of the byte segment
        if received == b'#':
            return None

        # Find out the chunk size and read the same amount of byte from the socket stream
        chunk_size = int(received.decode('utf-8'))
        return self.rfile.read(chunk_size)



from enum import Enum


class Operation(Enum):
    HELLO = '0.DOIP/Op.Hello'
    CREATE = '0.DOIP/Op.Create'
    RETRIEVE = '0.DOIP/Op.Retrieve'
    UPDATE = '0.DOIP/Op.Update'
    DELETE = '0.DOIP/Op.Delete'
    SEARCH = '0.DOIP/Op.Search'
    LIST_OPERATION = '0.DOIP/Op.ListOperations'


class ResponseStatus(Enum):
    SUCCESS = '0.DOIP/Status.001'
    INVALID = '0.DOIP/Status.101'
    UNAUTHENTICATED = '0.DOIP/Status.102'
    UNAUTHORIZED = '0.DOIP/Status.103'
    UNKNOWN_DO = '0.DOIP/Status.104'
    DUPLICATED_PID = '0.DOIP/Status.105'
    UNKNOWN_OPERATION = '0.DOIP/Status.200'
    UNKNOWN_ERROR = '0.DOIP/Status.500'


class SegmentType(Enum):
    JSON = 'JSON'
    BYTES = 'BYTES'
    EMPTY = 'EMPTY'



import json
from collections.abc import Iterator
from json import JSONDecodeError
from socketserver import BaseRequestHandler

from doip_sdk.constant import Operation
from doip_sdk.utils import SocketReader, send_unknown_operation_response, send_invalid_request_response


class DOIPHandler(BaseRequestHandler):
    """
    This is the base class for all DOIP-conformed handler. Whenever a new handler is implemented, it must inherit from
    this class. One do not create an instance of a handler directly, but rather pass the handler class to the
    :py:class:`DOIPServer` to create a DOIP server.
    """

    def hello(self, first_segment: dict, _: Iterator[bytearray]):
        send_unknown_operation_response(self.request, first_segment.get('requestId'))

    def create(self, first_segment: dict, _: Iterator[bytearray]):
        send_unknown_operation_response(self.request, first_segment.get('requestId'))

    def retrieve(self, first_segment: dict, _: Iterator[bytearray]):
        send_unknown_operation_response(self.request, first_segment.get('requestId'))

    def update(self, first_segment: dict, _: Iterator[bytearray]):
        send_unknown_operation_response(self.request, first_segment.get('requestId'))

    def delete(self, first_segment: dict, _: Iterator[bytearray]):
        send_unknown_operation_response(self.request, first_segment.get('requestId'))

    def search(self, first_segment: dict, _: Iterator[bytearray]):
        send_unknown_operation_response(self.request, first_segment.get('requestId'))

    def list_operations(self, first_segment: dict, _: Iterator[bytearray]):
        send_unknown_operation_response(self.request, first_segment.get('requestId'))

    def extended_operation(self, first_segment: dict, _: Iterator[bytearray]):
        send_unknown_operation_response(self.request, first_segment.get('requestId'))

    def handle(self):
        reader = SocketReader(self.request)
        chunks = reader.get_chunks()
        first_segment = next(chunks).decode('utf-8')

        if not first_segment.startswith('{'):
            send_invalid_request_response(socket=self.request, reason='The first segment is not a JSON object.')
            return

        try:
            first_segment = json.loads(first_segment)
        except JSONDecodeError:
            send_invalid_request_response(
                socket=self.request,
                reason='Cannot parse the JSON object from the first segment.'
            )
            return

        match first_segment['operationId']:
            case Operation.HELLO.value:
                self.hello(first_segment, chunks)
            case Operation.CREATE.value:
                self.create(first_segment, chunks)
            case Operation.UPDATE.value:
                self.update(first_segment, chunks)
            case Operation.DELETE.value:
                self.delete(first_segment, chunks)
            case Operation.SEARCH.value:
                self.search(first_segment, chunks)
            case Operation.LIST_OPERATION.value:
                self.list_operations(first_segment, chunks)
            case _:
                self.extended_operation(first_segment, chunks)



import ssl
from pathlib import Path
from socketserver import ThreadingTCPServer

from doip_sdk.base import DOIPHandler
from doip_sdk.selfsigned import generate_selfsigned_cert


class DOIPServer(ThreadingTCPServer):

    def __init__(self, host: str, port: int, handler: DOIPHandler, key_cert_files: tuple[str, str] | None = None,
                 bind_and_activate=True):
        super(DOIPServer, self).__init__((host, port), handler, bind_and_activate=False)

        # Auto generate a self-signed certificate, if not provided
        if key_cert_files:
            key_file, cert_file = key_cert_files
        else:
            key, cert = generate_selfsigned_cert(hostname=host)

            key_file = Path('ssl/key.pem')
            cert_file = Path('ssl/cert.pem')
            key_file.parent.mkdir(parents=True, exist_ok=True)

            with key_file.open('wb+') as f:
                f.write(key)
            with cert_file.open('wb+') as f:
                f.write(cert)

        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        context.load_cert_chain(certfile=cert_file, keyfile=key_file)
        self.socket = context.wrap_socket(self.socket, server_side=True)

        # Keep the paths so that we can delete them later
        self.key_file = key_file
        self.cert_file = cert_file

        if bind_and_activate:
            self.server_bind()
            self.server_activate()

    def _remove_certificate(self):
        # Remove self-signed certificate
        self.key_file.unlink(missing_ok=True)
        self.cert_file.unlink(missing_ok=True)
        self.key_file.parent.rmdir()

    def start(self):
        """Call this method to start the server."""
        try:
            self.serve_forever()
        except (KeyboardInterrupt, SystemExit):
            print('Shutting down the server')
        finally:
            self._remove_certificate()



__all__ = [
    'DOIPHandler',
    'Operation',
    'SegmentType',
    'ResponseStatus',
    'DOIPResponse',
    'write_string_segment',
    'write_empty_segment',
    'write_file_segment',
    'send_unknown_operation_response',
    'send_invalid_request_response',
    'SocketReader',
    'DOIPServer'
]

from doip_sdk.base import DOIPHandler
from doip_sdk.base_server import DOIPServer
from doip_sdk.constant import Operation, ResponseStatus, SegmentType
from doip_sdk.model.response import DOIPResponse
from doip_sdk.utils import write_string_segment, SocketReader, write_empty_segment, write_file_segment, \
    send_invalid_request_response, send_unknown_operation_response



from typing import Any

from pydantic import BaseModel

from doip_sdk.constant import ResponseStatus


class DOIPResponse(BaseModel):
    requestId: str | None = None
    status: ResponseStatus
    attributes: list = None
    output: Any | None = None
